/**************************************************************************
 *                                                                        *
 *   Copyright (c) Digital Equipment  Corporation,  1991.   All  Rights   *
 *   Reserved.  Unpublished rights reserved under the copyright laws of   *
 *   the United States.                                                   *
 *                                                                        *
 *   The software  contained  on  this  media  is  proprietary  to  and   *
 *   embodies   the   confidential   technology  of  Digital  Equipment   *
 *   Corporation.  Possession, use, duplication or dissemination of the   *
 *   software  and media is authorized only pursuant to a valid written   *
 *   license from Digital Equipment Corporation.                          *
 *                                                                        *
 *   RESTRICTED RIGHTS LEGEND Use, duplication, or  disclosure  by  the   *
 *   U.S.   Government  is  subject  to  restrictions  as  set forth in   *
 *   Subparagraph  (c)(1)(ii)  of  DFARS  252.227-7013,   or   in   FAR   *
 *   52.227-19, as applicable.                                            *
 *                                                                        *
 **************************************************************************/
#ifdef SCCS
static char sccsid[]="@(#)read_ucd.c	6.1 Stardent 91/08/09";
#endif
/*
			Copyright (c) 1989 by
			Stardent Computer Inc.
			All Rights Reserved
	
	This software comprises unpublished confidential information of
	Stardent Computer Inc. and may not be used, copied or made
	available to anyone, except in accordance with the license
	under which it is furnished.
	
	This file is under sccs control at Stardent in:
	@phobos/sccs3.p/avs/examples/s.read_ucd.c
	
*/
/*-----------------------------------------------------*
 *                                                     *
 *               ****  read_ucd  ****                  *
 *                                                     *
 * read a file (ascii or binary) in the ucd format.    *
 * if the file is binary then it is assumed it was     *
 * created using the ucd_write module.                 *
 *                                                     *
 *-----------------------------------------------------*/

#include <stdio.h>
#include <avs/flow.h>
#include <avs/ui.h>
#include <avs/geom.h>
#include <avs/ucd_defs.h>

#define buffer_size  200
#define  NUM_CELL_TYPES  8

extern char *AVSstatic;

typedef struct _ucd_stats {
 int size;
 } Ucd_Stats_Type;

typedef struct _ctype_struct {
  int id, mat, n, cell_type;
  } Ctype;

typedef struct _ntype_struct {
  int n, node_list[20];
  } Ntype;

typedef struct _model_stats {
  char node_data_labels[100], cell_data_labels[100], model_data_labels[100],
       node_data_units[100], cell_data_units[100], model_data_units[100];

  int num_nodes, num_cells, num_node_data, num_cell_data, num_model_data,
      node_active_list[20], cell_active_list[20], model_active_list[20],
      num_node_comp, node_comp_list[20], num_cell_comp, cell_comp_list[20],
      num_model_comp, model_comp_list[20], num_nlist_nodes;
  } Mtype;


read_ucd (output, file_name, cell_con, td1, contour, td2, cell_sel, td3, 
	  model_sel)  

  char *file_name, *td1, *contour, *td2, *td3, *model_sel;

  UCD_structure **output;

  int cell_con;
  {

  char *tl, data_labels[200], string[40], line[buffer_size], cell_type[10], 
       *model_name, data_units[200], tmp_str[80];

  FILE *fp;

  float dx, dy, dz, x, y, z, xmin, xmax, ymin, ymax, zmin, zmax, min_extent[3], 
        max_extent[3];

  int node, num_mats, node_id, cell, n, cell_id, mat_id, i, node_list[40], 
      util_flag, data_len, j, *cell_list, ucd_flags, cell_tsize, node_csize; 

  Ctype *cells;

  float extent[6], *xc, *yc, *zc, *node_data, *cell_data, *model_data, 
        *min_node_data, *max_node_data, value, *min_cell_data,
        *max_cell_data, *min_model_data, *max_model_data;

  int active_list[10], comp_list[10], num_comp, 
      num_nodes, num_cells, num_node_data, num_cell_data, num_model_data,
      *bcell_nlists;

  Mtype model_stats;

  Ntype *cell_nlists;

  static int clist[] = {1, 2, 3, 4}; 

  Ucd_Stats_Type *ucd_stats;

  UCD_structure *model;

 /**************
  ***  body  ***
  **************/

  if (!file_name) 
    return(1);

  if (AVSstatic == NULL) {
     ucd_stats = (Ucd_Stats_Type *)malloc(sizeof(Ucd_Stats_Type));
     AVSstatic = (char *) ucd_stats;
  }
  else 
     ucd_stats = (Ucd_Stats_Type *)AVSstatic;

  if (AVSparameter_changed("read file", 0)) {
    if (*output) UCDstructure_free (*output);

    cell_nlists = NULL;
    bcell_nlists = NULL;


    /*  check to see if file is binary or ascii.  */

    if (check_file(file_name)) {
      if (!read_ascii (file_name, &cells, &cell_nlists, &model_stats, &xc, &yc, 
		       &zc, &node_data, &cell_data, &model_data, &min_node_data, 
		       &max_node_data, &min_cell_data, &max_cell_data, 
		       &min_model_data, &max_model_data)) {
        AVSerror (" Error in read_ucd: can't open file. \n");
        return (0);
        }
      }
    else {
      if (!read_binary (file_name, &cells, &bcell_nlists, &model_stats, &xc, 
			&yc, &zc, &node_data, &cell_data, &model_data, 
			&min_node_data, &max_node_data, &min_cell_data, 
			&max_cell_data, &min_model_data, &max_model_data)) {
        AVSerror (" Error in read_ucd: can't open bin file. \n");
        return (0);
        }
      }


    /*   set the model statistics: 

	   num_nodes - the number of nodes in the model.

	   num_cells - the number of cells (elements) in the model.

	   num_node_data - the total number of data values per node. if you
			   wanted to store three scalars and one 3-vector per
			   node then num_node_data would be set to six.
    
           num_cell_data - the total number of data values per cell. 

	   num_model_data - model data is used for values which apply to the
			    the model as a whole, not just nodes or cells. 
			    for example, the locations of loads could be stored
			    here.                                               


									       */
    util_flag = 0; 
    num_nodes = model_stats.num_nodes;
    num_cells = model_stats.num_cells;
    num_node_data = model_stats.num_node_data;
    num_cell_data = model_stats.num_cell_data;
    num_model_data = model_stats.num_model_data;
    strcpy (tmp_str, file_name);

    model_name = (char *)(strrchr(tmp_str, '/'));
    
    if (model_name) model_name++;  /* go to the next character */

    for (i = strlen(model_name); model_name[i] != '.'; i--);
    model_name[i] = '\0';
  

    /*  'cell_tsize' is the size of all of the cell topology (number of 
	nodes per cell) lists. it is used to allocate space for the node 
	lists for each cell. if all the cells were hexahedra (without 
	mid-edge nodes) then this would be set to num_cells*eight, if all 
	the cells were tetrahedra (without mid-edge nodes) then this would 
	be set to num_cells*four.                      
	
	'node_csize' is the size all of the node connectivity lists (the list
	of cells connected to a node). for a model constructed from all
	hexahedra this would be less than num_cells*eight. 

	both 'cell_tsize' and 'node_csize' are computed when the model is
	read (in the read_ascii or the read_bin functions) and are always
	equal. 
	
	'ucd_flags' specifies how labels for cells and nodes will be stored
	and whether material id's and cell types will be stored for each cell. 
	the following is a list of flags which can be or'ed together:
	
	  UCD_MATERIAL_IDS  - material id's for each cell will be stored.
	  UCD_NODE_NAMES    - node names will be stored.
	  UCD_CELL_NAMES    - cell names will be stored.
	  UCD_CELL_TYPES    - user defined cell type will be stored.
	  UCD_MID_EDGES     - cells will have mid-side edges.
	  UCD_CHAR          -  node/cell labels are character strings.
	  UCD_INT           -  node/cell labels are integers.                */ 

    cell_tsize = node_csize = model_stats.num_nlist_nodes;
    ucd_stats->size = cell_tsize;

    ucd_flags = UCD_INT | UCD_MATERIAL_IDS | UCD_NODE_NAMES | UCD_CELL_NAMES |
		UCD_CELL_TYPES | UCD_MID_EDGES;

    *output = (UCD_structure *)UCDstructure_alloc (model_name, 
              num_model_data, ucd_flags, num_cells, cell_tsize, 
	      num_cell_data, num_nodes, node_csize, num_node_data, util_flag);


    /*  store nodal coordinates.  */

    for (node = 0; node < num_nodes; node++) {
      x = xc[node], y = yc[node], z = zc[node];

      if (node) {
        xmin = (x < xmin ? x : xmin), xmax = (x > xmax ? x : xmax);
        ymin = (y < ymin ? y : ymin), ymax = (y > ymax ? y : ymax);
        zmin = (z < zmin ? z : zmin), zmax = (z > zmax ? z : zmax);
        }
      else {
        xmin = xmax = x;
        ymin = ymax = y;
        zmin = zmax = z;
        }

      if (!UCDnode_set_information (*output, node, node + 1, 0, clist)) {
        AVSerror ("Error in read_ucd: can't set node %d info.\n", node); 
        return (0);
        }
      }
  
    UCDstructure_set_node_positions (*output, xc, yc, zc);
  
  
    /*  store the model's extent (max/min dimensions).  */
  
    min_extent[0] = xmin, min_extent[1] = ymin, min_extent[2] = zmin;
    max_extent[0] = xmax, max_extent[1] = ymax, max_extent[2] = zmax;
  
    UCDstructure_set_extent (*output, min_extent, max_extent);
  
  
    /*  store cell type, topology.  */
  
    if (bcell_nlists) {  /*  if read in as binary file.  */
      for (cell = 0, j = 0; cell < num_cells; cell++, j += n) {
        n = cells[cell].n;

        add_cell (output, cell, cells[cell].cell_type, n, &bcell_nlists[j], 
		  cells[cell].mat);
	}
      }
    else
      for (cell = 0; cell < num_cells; cell++) 
        add_cell (output, cell, cells[cell].cell_type, cell_nlists[cell].n, 
		  cell_nlists[cell].node_list, cells[cell].mat);
  
  
    /*  store nodal data.  

	data at nodes is separated into components. successive nodes of a
	component are stored contiguously. for example, if you wish to
	store two scalars and one 3-vector then this would be three
	components and would be stored as (given n nodes in the model):

	0     node 0, component 1
	1     node 1, component 1
	           .
	           .
	           .
	n     node n, component 1
	n+1   node 0, component 2
	n+2   node 1, component 2
	           .
	           .
	           .
	2n    node n, component 2
	2n+1  node 0, component 3, 1st vector component
	2n+2  node 0, component 3, 2nd vector component
	2n+3  node 0, component 3, 3rd vector component
	2n+4  node 1, component 3, 1st vector component
	2n+5  node 1, component 3, 2nd vector component
	2n+6  node 1, component 3, 3rd vector component
	           .
	           .
	           .
	5n-2  node n, component 3, 1st vector component
	5n-1  node n, component 3, 2nd vector component
	5n    node n, component 3, 3rd vector component


        the node component list has an entry for each compoment. each 
	entry specifies the dimensionality for each component. for example,
	a scalar component would have an entry of one and a 3-vector would
	have an entry of three.

        the node active list can be used to specify which data type is 
        currently active (being used to display results). the list has
        an entry for each data component. if that entry is not zero
        then it is active. currently, this functionality is not used by 
        any modules.                                                      

	node labels specify the user defined name of each data component.

	node units specify the user defined units of each data component.

	node minmax specifies the maximum and minimum values for each
	data component. this has no meaning for vector components although
	the max/min magnitude of the vector could be used.
	
	                                                                   */
  
    if (num_node_data) {
      UCDstructure_set_node_components (*output, model_stats.node_comp_list, 
					model_stats.num_node_comp);


      UCDstructure_set_node_active (*output, model_stats.node_active_list);

      UCDstructure_set_node_labels (*output, model_stats.node_data_labels, ".");

      UCDstructure_set_node_units (*output, model_stats.node_data_units, ".");

      if (model_stats.num_node_comp > 1) {
        for (i = 0; model_stats.node_data_labels[i] != '.'; i++)
	  string[i] = model_stats.node_data_labels[i];
        string[i] = '\0';
        }
      else
	strcpy (string, model_stats.node_data_labels);

      AVSmodify_parameter ("Node Type", AVS_MINVAL | AVS_VALUE, string, 
		           model_stats.node_data_labels, ".");
  
      UCDstructure_set_node_minmax (*output, min_node_data, max_node_data);
  
      if (!UCDstructure_set_node_data (*output, node_data)) {
        AVSerror ("Error in read_ucd: can't set node data.\n"); 
        return (0);
        }
      }
    else 
      AVSmodify_parameter ("Node Type", AVS_MINVAL | AVS_VALUE, " ", 
		           "<no data>", ".");
  
  
    /*  store cell data.  */
  
    if (num_cell_data) {
      UCDstructure_set_cell_components (*output, model_stats.cell_comp_list, 
					model_stats.num_cell_comp);

      UCDstructure_set_cell_active (*output, model_stats.cell_active_list);

      UCDstructure_set_cell_labels (*output, model_stats.cell_data_labels, ".");

      for (i = 0; model_stats.cell_data_labels[i] != '.'; i++)
	string[i] = model_stats.cell_data_labels[i];
      string[i] = '\0';

      AVSmodify_parameter ("Cell Type", AVS_MINVAL | AVS_VALUE, string, 
		           model_stats.cell_data_labels, ".");
  
      if (!UCDstructure_set_cell_data (*output, cell_data)) {
        AVSerror ("Error in read_ucd: can't set cell data.\n"); 
        return (0);
        }
      }
    else {
      AVSmodify_parameter ("Cell Type", AVS_MINVAL | AVS_VALUE, " ", 
		           "<no data>", ".");
      }
  
  
    /*  store model data.  */
  
    if (num_model_data) {
      UCDstructure_set_data_labels (*output, model_stats.model_data_labels, ".");

      for (i = 0; model_stats.model_data_labels[i] != '.'; i++)
	string[i] = model_stats.model_data_labels[i];
      string[i] = '\0';

      AVSmodify_parameter ("Model Type", AVS_MINVAL | AVS_VALUE, string, 
		           model_stats.model_data_labels, ".");

      if (!UCDstructure_set_data (*output, model_data)) {
        AVSerror ("Error in read_ucd: can't set model data.\n"); 
        return (0);
        }
      }
    else 
      AVSmodify_parameter ("Model Type", AVS_MINVAL | AVS_VALUE, " ", 
		           "<no data>", ".");

    if (*output) {
      free (xc);
      free (yc);
      free (zc);

      if (num_node_data) {
	free (node_data);
        free (min_node_data);
        free (max_node_data);
        }

      if (num_cell_data) {
	free (cell_data);
        free (min_cell_data);
        free (max_cell_data);
	}

      if (num_model_data) {
	free (model_data);
        free (min_model_data);
        free (max_model_data);
	}

      free (cells);

      if (cell_nlists) 
	free (cell_nlists);

      if (bcell_nlists) 
	free (bcell_nlists);
      }
    }
  return(1);
  }
  
  
/*-----------------------------------------------------*
 *                                                     *
 *            ****  read_data  ****                    *
 *                                                     *
 * read a block of data from an ascii file.            *
 *-----------------------------------------------------*/

read_data (fp, num_data, data, min_data, max_data, data_labels, units,
	   string, num_comp, comp_list)

  char *data_labels, *units, *string;

  FILE *fp;

  float *data, *min_data, *max_data;

  int num_data, *num_comp, *comp_list;
  {

  char line[80];

  float value;

  int offset, num_vals, node, id, j, i, n, m;

 /**************
  ***  body  ***
  **************/

  /*  read the number of components and their dimension.  */

  fscanf (fp, "%d", num_comp);

  for (i = 0; i < *num_comp; i++) 
    fscanf (fp, "%d", &comp_list[i]);

  fgets (line, buffer_size, fp);


  /*  read the component names and units.  */

  for (i = 0, n = 0, m = 0; i < *num_comp; i++) {
    fgets (line, buffer_size, fp);
    if (i == 0) strcpy (string, line);

    for (j = 0; line[j] == ' '; j++);

    for (; line[j] != ','; j++)
      data_labels[n++] = line[j];

    for (++j; line[j] == ' '; j++);

    for (; line[j] != '\n'; j++)
      units[m++] = line[j];

    if (i != *num_comp - 1) {
      data_labels[n++] = '.';
      units[m++] = '.';
      }
    }

  data_labels[n] = '\0';
  units[m] = '\0';

  for (i = 0; i < *num_comp; i++) 
    min_data[i] = max_data[i] = 0.0;

  for (node = 0; node < num_data; node++) {
    fscanf (fp, "%d", &id);
    offset = 0;

    for (i = 0; i < *num_comp; i++) {
      num_vals = comp_list[i];

      for (j = 0; j < num_vals; j++) {
        fscanf (fp, "%f",  &value);
        data[node * num_vals + offset + j] = value;
	}


      /*  only set max/min for scalar data.  */

      if (num_vals == 1) {
        if (node == 0) 
          min_data[i] = max_data[i] = value;
        else {
	  if (min_data[i] > value) min_data[i] = value;
	  if (max_data[i] < value) max_data[i] = value;
	  }
	}

      offset += num_data * num_vals;
      }
    }
  }


/*-----------------------------------------------------*
 *                                                     *
 *               ****  check_file  ****                *
 *                                                     *
 * check to see if the file to read is binary or not.  *
 *-----------------------------------------------------*/

check_file (file_name)
  char *file_name;
  { 

  char c;

  FILE *fp;

  int n;

 /**************
  ***  body  ***
  **************/

  if (!(fp = fopen (file_name, "r"))) {
    AVSerror (" Error in read_ucd: can't read file. \n");
    return (0);
    }

  c = fgetc(fp); 

  if (c == 0x5) {
    fclose (fp);
    return (0);
    }

  fclose (fp);
  return (1);
  }


/*-----------------------------------------------------*
 *                                                     *
 *               ****  add_cell  ****                  *
 *                                                     *
 * add a cell topology to the ucd structure.           *
 *-----------------------------------------------------*/

add_cell (model, cell, ucd_cell_type, n, rnode_list, mat_id) 

  int cell, ucd_cell_type, n, *rnode_list, mat_id;

  UCD_structure **model;
  {

  static char *cell_type[] = {"pt", "line", "tri", "quad", "tet",
			      "pyr", "prism", "hex"};

  int i, j, num_me_nodes, num_nodes, cell_found, me_flag, node_list[40];

 /**************
  ***  body  ***
  **************/

  if (ucd_cell_type < NUM_CELL_TYPES) {
    me_flag = 0;
    num_nodes = UCD_num_nodes[ucd_cell_type];


    /*  if the number of nodes read (n) is equal to the number of
	nodes without mid-side nodes then send the read node list.  
        else set the mid-edge flag to indicate which edges have 
	mide-edge nodes.                                            */

    for (i = 0; i < num_nodes; i++) 
      node_list[i] = rnode_list[i] - 1;

    if (n != num_nodes) {
      for (i = num_nodes, num_me_nodes = 0; i < n; i++)
        if (rnode_list[i] != 0) {
	  node_list[num_nodes + num_me_nodes] = rnode_list[i] - 1;
	  me_flag = me_flag | (0x1 << (i - num_nodes)); 
	  num_me_nodes++;
	  }
      }

    UCDcell_set_information (*model, cell, cell + 1, cell_type[ucd_cell_type], 
			     mat_id - 1, ucd_cell_type, me_flag, node_list); 
    }
  }


/*-----------------------------------------------------*
 *                                                     *
 *               ****  read_ascii  ****                *
 *                                                     *
 * read a ucd ascii file.                              *
 *-----------------------------------------------------*/

read_ascii (file_name, pcells, pcell_nlists, model, pxc, pyc, pzc, pnode_data, 
	    pcell_data, pmodel_data, pmin_node_data, pmax_node_data, 
	    pmin_cell_data, pmax_cell_data, pmin_model_data, pmax_model_data)

  char *file_name;

  Ctype **pcells;

  float **pxc, **pyc, **pzc, **pnode_data, **pcell_data, **pmodel_data,
        **pmin_node_data, **pmax_node_data, **pmin_cell_data, 
	**pmax_cell_data, **pmin_model_data, **pmax_model_data;

  Mtype *model;

  Ntype **pcell_nlists;

  {
  char *tl, node_data_labels[200], string[40], line[buffer_size], cell_type[10], 
       *model_name, cell_data_labels[200], model_data_labels[200];

  Ctype *cells;

  FILE *fp;

  float *xc, *yc, *zc, *node_data, *cell_data, *model_data, x, y, z,
        *min_node_data, *max_node_data, *min_cell_data, *max_cell_data,
        *min_model_data, *max_model_data;

  int node, node_id, cell, n, num_nodes, num_cells, num_node_data, 
      num_cell_data, num_model_data, cell_id, mat_id, i, node_list[40], 
      util_flag, data_len, j, node_active_list[20], cell_tsize, node_csize, 
      node_comp_list[20], num_node_comp, cell_active_list[20], 
      cell_comp_list[20], num_cell_comp, model_active_list[20], 
      model_comp_list[20], num_model_comp, ucd_cell_type, num_nlist_nodes;

  Ntype *cell_nlists;

  static char *cell_list[] = {"pt", "line", "tri", "quad", "tet",
			       "pyr", "prism", "hex"};

 /**************
  ***  body  ***
  **************/

  node_data = cell_data = model_data = NULL;

  if (!(fp = fopen (file_name, "r")))
    return (0);

  fgets (line, buffer_size, fp);


  /*  skip comments.  */

  while (line[0] == '#') 
    fgets (line, buffer_size, fp);


  /*  read model stats.  */ 

  sscanf (line, "%d %d %d %d %d", &num_nodes, &num_cells, &num_node_data, 
          &num_cell_data, &num_model_data);

  cells = (Ctype *)malloc(sizeof(Ctype) * num_cells);
  cell_nlists = (Ntype *)malloc(sizeof(Ntype) * num_cells);
  num_nlist_nodes = 0;

  xc = (float *)malloc(sizeof(float) * num_nodes);
  yc = (float *)malloc(sizeof(float) * num_nodes);
  zc = (float *)malloc(sizeof(float) * num_nodes);


  /*  read nodal coordinates.  */

  for (node = 0; node < num_nodes; node++) {
    fscanf (fp, "%d %f %f %f", &node_id, &x, &y, &z); 
    xc[node] = x, yc[node] = y, zc[node] = z;
    }
  
  
  /*  read cell type, topology.  */
  
  fgets (line, buffer_size, fp);
  
  for (cell = 0; cell < num_cells; cell++) {
    fgets (line, buffer_size, fp);
    tl = line; 
    n = 0;
  
    get_token (&tl, &cells[cell].id, "%d");
    get_token (&tl, &cells[cell].mat, "%d");
  
    i = get_token (&tl, cell_type, "%s");
    cell_type[i] = '\0';

    for (ucd_cell_type = 0; ucd_cell_type < NUM_CELL_TYPES; ucd_cell_type++)
      if (!strcmp(cell_list[ucd_cell_type], cell_type))
        break;

    if (ucd_cell_type < NUM_CELL_TYPES) 
      cells[cell].cell_type = ucd_cell_type;
    else {
      AVSerror (" Error in read_ucd: unknown cell type. \n");
      cells[cell].cell_type = 0;
      }

    while (get_token(&tl, &cell_nlists[cell].node_list[n++], "%d"));
    cell_nlists[cell].n = cells[cell].n = n - 1;
    num_nlist_nodes += n - 1;
    }
  
  
  /*  read nodal data.  */
  
  if (num_node_data) {
    node_data = (float *)malloc(sizeof(float) * num_nodes * num_node_data);
    min_node_data = (float *)malloc(sizeof(float) * num_node_data);
    max_node_data = (float *)malloc(sizeof(float) * num_node_data);

    read_data (fp, num_nodes, node_data, min_node_data, max_node_data, 
	       model->node_data_labels, model->node_data_units, string, 
	       &model->num_node_comp, model->node_comp_list);

    model->node_active_list[0] = 1;

    for (i = 1; i < model->num_node_comp; i++)
      model->node_active_list[i] = 0;
  
    if (model->num_node_comp > 1) {
      for (i = 0; model->node_data_labels[i] != '.'; i++)
        string[i] = model->node_data_labels[i];
      string[i] = '\0'; 
      }
    else 
      strcpy (string, model->node_data_labels);
    }
  
  
  /*  read cell data.  */
  
  if (num_cell_data) {
    cell_data = (float *)malloc(sizeof(float) * num_cells * num_cell_data);
    min_cell_data = (float *)malloc(sizeof(float) * num_cell_data);
    max_cell_data = (float *)malloc(sizeof(float) * num_cell_data);
  
    read_data (fp, num_cells, cell_data, min_cell_data, max_cell_data, 
	       model->cell_data_labels, model->cell_data_units,
	       string, &model->num_cell_comp, model->cell_comp_list);

    model->cell_active_list[0] = 1;

    for (i = 1; i < model->num_cell_comp; i++)
      model->cell_active_list[i] = 0;
  
    if (model->num_cell_comp > 1) {
      for (i = 0; model->cell_data_labels[i] != '.'; i++)
        string[i] = model->cell_data_labels[i];
      string[i] = '\0'; 
      }
    else 
      strcpy (string, model->cell_data_labels);
    }
  
  
  /*  read and store model data.  */
  
  if (num_model_data) {
    model_data = (float *)malloc(sizeof(float) * num_model_data);
    min_model_data = (float *)malloc(sizeof(float) * num_model_data);
    max_model_data = (float *)malloc(sizeof(float) * num_model_data);
  
    read_data (fp, num_model_data, model_data, min_model_data, max_model_data, 
	       model->model_data_labels, model->model_data_units,
	       string, &model->num_model_comp, model->model_comp_list);

    if (model->num_model_comp > 1) {
      for (i = 0; model->model_data_labels[i] != '.'; i++)
        string[i] = model->model_data_labels[i];
      string[i] = '\0'; 
      }
    else 
      strcpy (string, model->model_data_labels);
    }

  *pxc = xc, *pyc = yc, *pzc = zc;
  *pcells = cells;
  *pnode_data = node_data;
  *pcell_data = cell_data;
  *pcell_nlists = cell_nlists;
  *pmodel_data = model_data;

  model->num_nodes = num_nodes;
  model->num_cells = num_cells;
  model->num_nlist_nodes = num_nlist_nodes;
  model->num_node_data = num_node_data;
  model->num_cell_data = num_cell_data;
  model->num_model_data = num_model_data;

  *pmax_node_data = max_node_data;
  *pmin_node_data = min_node_data;
  *pmax_cell_data = max_cell_data;
  *pmin_cell_data = min_cell_data;
  *pmax_model_data = max_model_data;
  *pmin_model_data = min_model_data;

  fclose (fp);

  return(1);
  }
  

/*-----------------------------------------------------*
 *                                                     *
 *               ****  read_binary  ****               *
 *                                                     *
 * read a ucd binary file.                             *
 *-----------------------------------------------------*/

read_binary (file_name, pcells, pcell_nlists, model, pxc, pyc, pzc, pnode_data,
	     pcell_data, pmodel_data, pmin_node_data, pmax_node_data, 
	     pmin_cell_data, pmax_cell_data, pmin_model_data, pmax_model_data)

  char *file_name;

  float **pxc, **pyc, **pzc, **pnode_data, **pcell_data, **pmodel_data,
        **pmin_node_data, **pmax_node_data, **pmin_cell_data, 
	**pmax_cell_data, **pmin_model_data, **pmax_model_data;

  int **pcell_nlists;

  Mtype *model;

  Ctype **pcells;
  {

  char *tl, node_data_labels[200], string[40], line[buffer_size], cell_type[10], 
       *model_name, cell_data_labels[200], model_data_labels[200], magic;

  Ctype *cells;

  FILE *fp;

  float *xc, *yc, *zc, *node_data, *cell_data, *model_data, x, y, z,
        *min_node_data, *max_node_data, *min_cell_data, *max_cell_data,
        *min_model_data, *max_model_data;

  int node, node_id, cell, n, num_nodes, num_cells, num_node_data, 
      num_cell_data, num_model_data, cell_id, mat_id, i, node_list[40], 
      util_flag, data_len, j, *cell_list, node_active_list[20], 
      node_comp_list[20], num_node_comp, cell_active_list[20], 
      cell_comp_list[20], num_cell_comp, model_active_list[20], 
      model_comp_list[20], num_model_comp, *cell_nlists, num_nlist_nodes;

 /**************
  ***  body  ***
  **************/

  node_data = cell_data = model_data = NULL;

  if (!(fp = fopen (file_name, "r")))
    return (0);

  fread ((char *)&magic, sizeof(char), 1, fp);
  fread ((char *)model, sizeof(Mtype), 1, fp);

  num_nodes = model->num_nodes;
  num_cells = model->num_cells;
  num_node_data = model->num_node_data;
  num_cell_data = model->num_cell_data;
  num_model_data = model->num_model_data;
  num_nlist_nodes = model->num_nlist_nodes;

  cells = (Ctype *)malloc(sizeof(Ctype) * num_cells);
  cell_nlists = (int *)malloc(sizeof(int) * num_nlist_nodes);

  fread  ((char *)cells, sizeof(Ctype), num_cells, fp);
  fread  ((char *)cell_nlists, sizeof(int), num_nlist_nodes, fp);

  xc = (float *)malloc(sizeof(float) * num_nodes);
  yc = (float *)malloc(sizeof(float) * num_nodes);
  zc = (float *)malloc(sizeof(float) * num_nodes);

  fread ((char *)xc, sizeof(float), num_nodes, fp);
  fread ((char *)yc, sizeof(float), num_nodes, fp);
  fread ((char *)zc, sizeof(float), num_nodes, fp);

  if (num_node_data) {
    node_data = (float *)malloc(sizeof(float) * num_nodes * num_node_data);
    min_node_data = (float *)malloc(sizeof(float) * num_node_data);
    max_node_data = (float *)malloc(sizeof(float) * num_node_data);

    fread ((char *)min_node_data, sizeof(float), num_node_data, fp);
    fread ((char *)max_node_data, sizeof(float), num_node_data, fp);
    fread ((char *)node_data, sizeof(float), num_node_data * num_nodes, fp);
    }

  if (num_cell_data) {
    cell_data = (float *)malloc(sizeof(float) * num_cells * num_cell_data);
    min_cell_data = (float *)malloc(sizeof(float) * num_cell_data);
    max_cell_data = (float *)malloc(sizeof(float) * num_cell_data);

    fread ((char *)min_cell_data, sizeof(float), num_cell_data, fp);
    fread ((char *)max_cell_data, sizeof(float), num_cell_data, fp);
    fread ((char *)cell_data, sizeof(float), num_cell_data, fp);
    }

  if (num_model_data) {
    model_data = (float *)malloc(sizeof(float) * num_model_data);
    min_model_data = (float *)malloc(sizeof(float) * num_model_data);
    max_model_data = (float *)malloc(sizeof(float) * num_model_data);

    fread ((char *)min_model_data, sizeof(float), num_model_data, fp);
    fread ((char *)max_model_data, sizeof(float), num_model_data, fp);
    fread ((char *)model_data, sizeof(float), num_model_data, fp);
    }

  *pxc = xc, *pyc = yc, *pzc = zc;
  *pcells = cells;
  *pcell_nlists = cell_nlists; 
  *pnode_data = node_data;
  *pcell_data = cell_data;
  *pmodel_data = model_data;

  model->num_nodes = num_nodes;
  model->num_cells = num_cells;
  model->num_node_data = num_node_data;
  model->num_cell_data = num_cell_data;
  model->num_model_data = num_model_data;

  *pmax_node_data = max_node_data;
  *pmin_node_data = min_node_data;
  *pmax_cell_data = max_cell_data;
  *pmin_cell_data = min_cell_data;
  *pmax_model_data = max_model_data;
  *pmin_model_data = min_model_data;

  fclose (fp);

  return(1);
  }
  

/*-----------------------------------------------------*
 *                                                     *
 *               ****  get_token  ****                 *
 *                                                     *
 * returns the next value in tline[] separated by a    *
 * blank.                                              *
 *-----------------------------------------------------*/

get_token (tline, data, control)
  char **tline, *data, *control;
  {

  char *line, token[80], c;

  int ntc;

 /**************
  ***  body  ***
  **************/

  c = 0;
  line = *tline;

  for (ntc = 0; c != '\n'; line++) {
    if (((c  = *line) != ' ') && (c != '\n'))
      token[ntc++] = c;
    else if (ntc != 0) {
      token[ntc] = '\0';
      sscanf (token, control, data);
      *tline = line;
      break;
      }
    }

  return (ntc);
  }


/*-----------------------------------------------------*
 *                                                     *
 *          ****  read_ucd_init  ****                  *
 *                                                     *
 *-----------------------------------------------------*/

read_ucd_init()
  {
  AVSstatic = (char *)0;
  }


/*-----------------------------------------------------*
 *                                                     *
 *          ****  read_ucd_finis  ****                 *
 *                                                     *
 *-----------------------------------------------------*/

read_ucd_finis()
  {
  Ucd_Stats_Type *ucd_stats;

  if (AVSstatic == NULL) return;

  free (AVSstatic);
  }



read_ucd_desc()
  {
  char *string, tmp[20], *init, *sep;

  int read_ucd(), param;

  static char *choices = "<data 1>.<data 2>.<data 3>.<data 4>.<data 5>";

 /**************
  ***  body  ***
  **************/

  AVSset_module_name ("Read ucd c", MODULE_DATA);

  AVScreate_output_port ("tet output", "ucd");

  param = AVSadd_parameter ("read file", "string", 0, 0, ".inp");
  AVSconnect_widget (param, "browser");

  AVSadd_parameter("Cell Connect", "boolean", 0, 0, 1);

  param = AVSadd_parameter("Node Data", "string", "Node Data", "Node Data", 
			   NULL);
  AVSconnect_widget(param, "text");

  AVSadd_parameter ("Node Type", "choice", "<data 1>", choices, ".");

  param = AVSadd_parameter("Cell Data", "string", "Cell Data", "Cell Data", 
			   NULL);
  AVSconnect_widget (param, "text");

  AVSadd_parameter ("Cell Type", "choice", "<data 1>", choices, ".");

  param = AVSadd_parameter("Model Data", "string", "Model Data", "Model Data", 
			   NULL);
  AVSconnect_widget (param, "text");

  AVSadd_parameter ("Model Type", "choice", "<data 1>", choices, ".");

  AVSset_init_proc (read_ucd_init);

  AVSset_destroy_proc (read_ucd_finis);

  AVSset_compute_proc (read_ucd);
  }

AVSinit_modules()
  {
  AVSmodule_from_desc (read_ucd_desc);
  }
